import java.io.Console;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.InputStream;
import java.security.KeyManagementException;
import java.security.KeyStore;
import java.security.KeyStoreException;
import java.security.NoSuchAlgorithmException;
import java.security.cert.CertificateException;
import java.util.UUID;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;
import java.util.concurrent.ThreadLocalRandom;
import java.util.concurrent.TimeUnit;
import java.util.concurrent.atomic.AtomicBoolean;

import org.infinispan.client.hotrod.RemoteCache;
import org.infinispan.client.hotrod.RemoteCacheManager;
import org.infinispan.client.hotrod.configuration.ConfigurationBuilder;

import javax.net.ssl.SSLContext;
import javax.net.ssl.TrustManagerFactory;
import javax.security.auth.callback.Callback;
import javax.security.auth.callback.CallbackHandler;
import javax.security.auth.callback.NameCallback;
import javax.security.auth.callback.PasswordCallback;
import javax.security.auth.callback.UnsupportedCallbackException;
import javax.security.sasl.AuthorizeCallback;
import javax.security.sasl.RealmCallback;

public class HotRodTestClient {

    RemoteCache cache;
    ExecutorService loadRunner;
    AtomicBoolean stopping;

    private RemoteCacheManager manager;
    private InfinispanCallbackHandler callbackHandler;

    public HotRodTestClient(String username, String password) throws NoSuchAlgorithmException, KeyStoreException, KeyManagementException {
        loadRunner = Executors.newFixedThreadPool(2);
        stopping = new AtomicBoolean(false);
        callbackHandler = new InfinispanCallbackHandler(username, password, "ApplicationRealm");
        createRemoteCacheManager();
    }

    public void createRemoteCacheManager() throws NoSuchAlgorithmException, KeyStoreException, KeyManagementException {
        ConfigurationBuilder cb = new ConfigurationBuilder();
        cb.withProperties(System.getProperties())
                .security().ssl().enable()
                .sslContext(createSSLContext())
                .security().authentication().enable()
                .saslMechanism("DIGEST-MD5")
                .serverName("localhostjdg")
                .callbackHandler(callbackHandler);

        manager = new RemoteCacheManager(cb.build());
        manager.start();
    }

    public void updateCache() {
        byte[] value = new byte[50];
        ThreadLocalRandom.current().nextBytes(value);
        UUID key = UUID.randomUUID();

        try {
	        if (cache.putIfAbsent(key, value) != null) {
	            // Key existed, ignore for test purpose since we are only interested in placing load on jdg.
	        }
        }catch(Exception e) {
        	e.printStackTrace();
        	throw e;
        }
        queuePut();
    }

    private void stop() throws InterruptedException {
        stopping.set(true);
        loadRunner.shutdown();
        loadRunner.awaitTermination(60, TimeUnit.SECONDS);
    }

    public void queuePut() {
        if (!stopping.get()) {
            loadRunner.submit(() -> {
                updateCache();
            });
        }
    }

    public static void main(String[] args) throws IOException, InterruptedException, NoSuchAlgorithmException, KeyStoreException, KeyManagementException {
        HotRodTestClient client = new HotRodTestClient(args[0], args[1]);
        client.initCache();
        client.queuePut();
        waitForUserInput();
        client.stop();
    }

    private void initCache() {
        cache = manager.getCache("testCache");
    }

    public static void waitForUserInput() throws IOException {
        System.out.println("Press enter to exit");
        System.in.read();
    }

    private SSLContext createSSLContext() throws NoSuchAlgorithmException, KeyStoreException, KeyManagementException {
        String trustfile = System.getProperty("infinispan.client.hotrod.trust_store_file_name");
        String trustpass = System.getProperty("infinispan.client.hotrod.trust_store_password");

        SSLContext context = SSLContext.getInstance("TLSv1.2");

        KeyStore trustStore = KeyStore.getInstance("JKS");
        try (InputStream stream = new FileInputStream(trustfile)) {
            trustStore.load(stream, trustpass.toCharArray());
        } catch (CertificateException e) {
            e.printStackTrace();
        } catch (FileNotFoundException e) {
            e.printStackTrace();
        } catch (IOException e) {
            e.printStackTrace();
        }

        TrustManagerFactory tmf = TrustManagerFactory.getInstance("SunX509");
        tmf.init(trustStore);

        context.init(null, tmf.getTrustManagers(), null);
        return context;
    }

    private class InfinispanCallbackHandler implements CallbackHandler {
        private final String username;
        private final String password;
        private final String realm;

        public InfinispanCallbackHandler(String username, String password, String realm) {
            this.username = username;
            this.password = password;
            this.realm = realm;
        }

        @Override
        public void handle(Callback[] callbacks) throws IOException, UnsupportedCallbackException {
            for (Callback callback : callbacks) {
                if (callback instanceof NameCallback) {
                    NameCallback nameCallback = (NameCallback) callback;
                    nameCallback.setName(username);
                } else if (callback instanceof PasswordCallback) {
                    PasswordCallback passwordCallback = (PasswordCallback) callback;
                    passwordCallback.setPassword(password.toCharArray());
                } else if (callback instanceof AuthorizeCallback) {
                    AuthorizeCallback authorizeCallback = (AuthorizeCallback) callback;
                    authorizeCallback.setAuthorized(authorizeCallback.getAuthenticationID().equals(
                            authorizeCallback.getAuthorizationID()));
                } else if (callback instanceof RealmCallback) {
                    RealmCallback realmCallback = (RealmCallback) callback;
                    realmCallback.setText(realm);
                } else {
                    throw new UnsupportedCallbackException(callback);
                }
            }
        }
    }
}
